/*
 *
 * Difficulty (Updated)
 *
 */

const events = require('events');
const utils = require('./utils');

////////////////////////////////////////////////////////////////////////////////

// Main RingBuffer Function
const RingBuffer = function(maxSize) {

  let data = [];
  let cursor = 0;
  let isFull = false;

  // Append to Ring Buffer
  this.append = function(x) {
    if (isFull) {
      data[cursor] = x;
      cursor = (cursor + 1) % maxSize;
    } else {
      data.push(x);
      cursor += 1;
      if (data.length === maxSize) {
        cursor = 0;
        isFull = true;
      }
    }
  };

  // Average Ring Buffer
  this.avg = function() {
    const sum = data.reduce((a, b) => a + b);
    return sum / (isFull ? maxSize : cursor);
  };

  // Calculate Size of Ring Buffer
  this.size = function() {
    return isFull ? maxSize : cursor;
  };

  // Clear Ring Buffer
  this.clear = function() {
    data = [];
    cursor = 0;
    isFull = false;
  };
};

// Main Difficulty Function
const Difficulty = function(port, difficulty) {

  const _this = this;
  _this.options = difficulty;

  let lastTs, timeBuffer;
  const variance = difficulty.variance;
  const bufferSize = difficulty.targetSample;

  // Update Difficulty on Share Submission
  this.updateDifficulty = function(client) {
    const ts = (Date.now() / 1000) | 0;

    if (!lastTs) {
      lastTs = ts;
      timeBuffer = new RingBuffer(bufferSize);
      return;
    }

    const sinceLast = ts - lastTs;
    timeBuffer.append(sinceLast);
    lastTs = ts;
    const avg = timeBuffer.avg();
    let dDiff = _this.options.targetTime / avg;

    if((timeBuffer.size() > 0) && (timeBuffer.size() < _this.options.targetSample)) {
        return;
    }
    if((dDiff < (1 + variance)) && (dDiff > (1 - variance))) {
        return;
    }

    const newDiff = utils.toFixed(client.difficulty * dDiff, 8);

    if(newDiff > _this.options.maximum) newDiff = _this.options.maximum;
    if(newDiff < _this.options.minimum) newDiff = _this.options.minimum;

    timeBuffer.clear();
    _this.emit('newDifficulty', client, newDiff);

  };

  // Manage Individual Clients
  this.manageClient = function(client) {
    const stratumPort = client.socket.localPort;
    if (stratumPort != port) {
      console.error('Handling a client which is not of this vardiff?');
    }
    client.on('submit', () => _this.updateDifficulty(client));
  };
};

module.exports = Difficulty;
Difficulty.prototype.__proto__ = events.EventEmitter.prototype;
